from __future__ import annotations

from typing import Any, Literal

from pydantic import Field, field_validator, model_validator

from ..base import APIModel

__all__ = ("Weapon", "WeaponDetail", "WeaponProperty", "WeaponRefinement", "WeaponStatModifier")


class WeaponProperty(APIModel):
    """Weapon's property."""

    type: str = Field(alias="propType")
    init_value: float = Field(alias="initValue")
    growth_type: str = Field(alias="type")


class WeaponStatModifier(APIModel):
    """Weapon's stat modifier."""

    base: float = Field(alias="Base")
    levels: dict[str, float] = Field(alias="Levels")


class WeaponRefinement(APIModel):
    """Weapon's refinement."""

    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    parameters: list[float] = Field(alias="ParamList")


class WeaponDetail(APIModel):
    """Genshin Impact weapon detail."""

    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    rarity: Literal[1, 2, 3, 4, 5] = Field(alias="Rarity")
    icon: str = Field(alias="Icon")

    stat_modifiers: dict[str, WeaponStatModifier] = Field(alias="StatsModifier")
    xp_requirements: dict[str, float] = Field(alias="XPRequirements")
    ascension: dict[str, dict[str, float]] = Field(alias="Ascension")
    refinments: dict[str, WeaponRefinement] = Field(alias="Refinement")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"


class Weapon(APIModel):
    """Genshin Impact weapon."""

    id: int  # This field is not present in the API response.
    icon: str
    rarity: Literal[1, 2, 3, 4, 5] = Field(alias="rank")
    description: str = Field(alias="desc")
    names: dict[Literal["EN", "CHS", "KR", "JP"], str]
    name: str = Field(None)  # This value of this field is assigned in post processing.

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"

    @model_validator(mode="before")
    def _transform_names(cls, values: dict[str, Any]) -> dict[str, Any]:
        values["names"] = {
            "EN": values.pop("EN"),
            "CHS": values.pop("CHS"),
            "KR": values.pop("KR"),
            "JP": values.pop("JP"),
        }
        return values



from __future__ import annotations

from typing import Any, Literal, Self

from pydantic import Field, computed_field, field_validator, model_validator

from ...utils import replace_placeholders
from ..base import APIModel

__all__ = (
    "Relic",
    "RelicSet",
    "RelicSetDetail",
    "RelicSetEffect",
    "RelicSetEffects",
    "SetDetailSetEffect",
)


class Relic(APIModel):
    """HSR relic."""

    id: int = Field(None)  # This field is not present in the API response.
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    story: str = Field(alias="Story")

    @computed_field
    @property
    def icon(self) -> str:
        """Relic's icon URL."""
        relic_id = str(self.id)[1:4]
        part_id = str(self.id)[-1]
        return f"https://api.hakush.in/hsr/UI/relicfigures/IconRelic_{relic_id}_{part_id}.webp"


class SetDetailSetEffect(APIModel):
    """Relic set detail's set effect."""

    description: str = Field(alias="Desc")
    parameters: list[float] = Field(alias="ParamList")

    @model_validator(mode="after")
    def _format_parameters(self) -> Self:
        self.description = replace_placeholders(self.description, self.parameters)
        return self


class SetDetailSetEffects(APIModel):
    """Relic set detail's set effects."""

    two_piece: SetDetailSetEffect
    four_piece: SetDetailSetEffect | None = None


class RelicSetDetail(APIModel):
    """HSR relic set detail."""

    name: str = Field(alias="Name")
    icon: str = Field(alias="Icon")
    parts: dict[str, Relic] = Field(alias="Parts")
    set_effects: SetDetailSetEffects = Field(alias="RequireNum")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        icon_id = value.split("/")[-1].split(".")[0]
        return f"https://api.hakush.in/hsr/UI/itemfigures/{icon_id}.webp"

    @field_validator("set_effects", mode="before")
    def _assign_set_effects(cls, value: dict[str, Any]) -> dict[str, Any]:
        return {"two_piece": value["2"], "four_piece": value.get("4")}

    @field_validator("parts", mode="before")
    def _convert_parts(cls, value: dict[str, Any]) -> dict[str, Any]:
        return {key: Relic(id=int(key), **value[key]) for key in value}


class RelicSetEffect(APIModel):
    """Relic set effect."""

    descriptions: dict[Literal["en", "cn", "kr", "jp"], str]
    description: str = Field(None)  # The value of this field is assigned in post processing.
    parameters: list[float] = Field(alias="ParamList")

    @model_validator(mode="before")
    def _assign_descriptions(cls, value: dict[str, Any]) -> dict[str, Any]:
        value["descriptions"] = {
            "en": value.pop("en"),
            "cn": value.pop("cn"),
            "kr": value.pop("kr"),
            "jp": value.pop("jp"),
        }
        return value


class RelicSetEffects(APIModel):
    """Relic set's set effects."""

    two_piece: RelicSetEffect
    four_piece: RelicSetEffect | None = None


class RelicSet(APIModel):
    """HSR relic set."""

    id: int  # This field is not present in the API response.
    icon: str
    names: dict[Literal["en", "cn", "kr", "jp"], str]
    name: str = Field(None)  # The value of this field is assigned in post processing.
    set_effect: RelicSetEffects = Field(alias="set")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        icon_id = value.split("/")[-1].split(".")[0]
        return f"https://api.hakush.in/hsr/UI/itemfigures/{icon_id}.webp"

    @field_validator("set_effect", mode="before")
    def _assign_set_effect(cls, value: dict[str, Any]) -> dict[str, Any]:
        return {"two_piece": value["2"], "four_piece": value.get("4")}

    @model_validator(mode="before")
    def _assign_names(cls, value: dict[str, Any]) -> dict[str, Any]:
        value["names"] = {
            "en": value.pop("en"),
            "cn": value.pop("cn"),
            "kr": value.pop("kr"),
            "jp": value.pop("jp"),
        }
        return value



from __future__ import annotations

from typing import Any, Literal

from pydantic import Field, computed_field, field_validator, model_validator

from ...constants import ZZZ_SAB_RARITY_CONVERTER
from ...enums import ZZZSpecialty
from ...utils import cleanup_text
from ..base import APIModel

__all__ = (
    "Weapon",
    "WeaponDetail",
    "WeaponLevel",
    "WeaponProp",
    "WeaponRefinement",
    "WeaponStar",
    "WeaponType",
)


class Weapon(APIModel):
    """ZZZ weapon (w-engine) model."""

    id: int
    icon: str
    name: str = Field(None)  # This field doesn't exist in the API response
    names: dict[Literal["EN", "JA", "CHS", "KO"], str]
    specialty: ZZZSpecialty = Field(alias="type")
    rarity: Literal["S", "A", "B"] = Field(alias="rank")

    @field_validator("rarity", mode="before")
    @classmethod
    def __convert_rarity(cls, value: int | None) -> Literal["S", "A", "B"] | None:
        return ZZZ_SAB_RARITY_CONVERTER[value] if value is not None else None

    @field_validator("icon")
    @classmethod
    def __convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/zzz/UI/{value}.webp"

    @model_validator(mode="before")
    @classmethod
    def __pop_names(cls, values: dict[str, Any]) -> dict[str, Any]:
        values["names"] = {
            "EN": values.pop("EN"),
            "KO": values.pop("KO"),
            "CHS": values.pop("CHS"),
            "JA": values.pop("JA"),
        }
        return values


class WeaponType(APIModel):
    """ZZZ weapon type model."""

    type: ZZZSpecialty
    name: str


class WeaponProp(APIModel):
    """ZZZ weapon property model."""

    name: str = Field(alias="Name")
    name2: str = Field(alias="Name2")
    format: str = Field(alias="Format")
    value: float = Field(alias="Value")

    @computed_field
    @property
    def formatted_value(self) -> str:
        """Formatted value of this prop."""
        if "%" in self.format:
            return f"{self.value / 100:.0%}%"
        return str(round(self.value))


class WeaponLevel(APIModel):
    """ZZZ weapon level model."""

    exp: int = Field(alias="Exp")
    rate: int = Field(alias="Rate")
    rate2: int = Field(alias="Rate2")


class WeaponStar(APIModel):
    """ZZZ weapon star model."""

    star_rate: int = Field(alias="StarRate")
    rand_rate: int = Field(alias="RandRate")


class WeaponRefinement(APIModel):
    """ZZZ weapon refinement model."""

    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")

    @field_validator("description")
    @classmethod
    def __cleanup_text(cls, value: str) -> str:
        return cleanup_text(value)


class WeaponDetail(APIModel):
    """ZZZ weapon (w-engine) detail model."""

    id: int = Field(alias="Id")
    code_name: str = Field(alias="CodeName")
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    description2: str = Field(alias="Desc2")
    short_description: str = Field(alias="Desc3")
    rarity: Literal["S", "A", "B"] | None = Field(alias="Rarity")
    icon: str = Field(alias="Icon")
    type: WeaponType = Field(alias="WeaponType")
    base_property: WeaponProp = Field(alias="BaseProperty")
    rand_property: WeaponProp = Field(alias="RandProperty")
    levels: dict[str, WeaponLevel] = Field(alias="Level")
    stars: dict[str, WeaponStar] = Field(alias="Stars")
    materials: str = Field(alias="Materials")
    refinements: dict[str, WeaponRefinement] = Field(alias="Talents")  # {'1': ..., '2': ...}
    """Dictionary of refinements, key starts from 1."""

    @field_validator("type", mode="before")
    @classmethod
    def __convert_type(cls, value: dict[str, str]) -> WeaponType:
        first_item = next(iter(value.items()))
        return WeaponType(type=ZZZSpecialty(int(first_item[0])), name=first_item[1])

    @field_validator("icon")
    @classmethod
    def __convert_icon(cls, value: str) -> str:
        value = value.split("/")[-1].split(".")[0]
        return f"https://api.hakush.in/zzz/UI/{value}.webp"

    @field_validator("rarity", mode="before")
    @classmethod
    def __convert_rarity(cls, value: int | None) -> Literal["S", "A", "B"] | None:
        return ZZZ_SAB_RARITY_CONVERTER[value] if value is not None else None



from __future__ import annotations

import pytest

import hakushin


@pytest.mark.asyncio
async def test_not_found_error() -> None:
    with pytest.raises(hakushin.errors.NotFoundError):
        async with hakushin.HakushinAPI(hakushin.Game.GI) as client:
            await client.fetch_character_detail("0")
        async with hakushin.HakushinAPI(hakushin.Game.HSR) as client:
            await client.fetch_character_detail(0)
        async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
            await client.fetch_character_detail(0)



from __future__ import annotations

import pytest

import hakushin


@pytest.mark.asyncio
async def test_new() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.HSR) as client:
        await client.fetch_new()


@pytest.mark.asyncio
async def test_characters() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.HSR) as client:
        await client.fetch_characters()


@pytest.mark.asyncio
async def test_character() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.HSR) as client:
        new = await client.fetch_new()
        for chara_id in new.character_ids:
            await client.fetch_character_detail(chara_id)


@pytest.mark.asyncio
async def test_light_cones() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.HSR) as client:
        await client.fetch_light_cones()


@pytest.mark.asyncio
async def test_light_cone() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.HSR) as client:
        hsr_new = await client.fetch_new()
        for light_cone_id in hsr_new.light_cone_ids:
            await client.fetch_light_cone_detail(light_cone_id)


@pytest.mark.asyncio
async def test_relic_set() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.HSR) as client:
        hsr_new = await client.fetch_new()
        for relic_set_id in hsr_new.relic_set_ids:
            await client.fetch_relic_set_detail(relic_set_id)



from __future__ import annotations

import pytest

import hakushin


@pytest.mark.asyncio
async def test_new() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.GI) as client:
        await client.fetch_new()


@pytest.mark.asyncio
async def test_characters() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.GI) as client:
        await client.fetch_characters()


@pytest.mark.asyncio
async def test_character() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.GI) as client:
        new = await client.fetch_new()
        for chara_id in new.character_ids:
            await client.fetch_character_detail(str(chara_id))


@pytest.mark.asyncio
async def test_weapons() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.GI) as client:
        await client.fetch_weapons()


@pytest.mark.asyncio
async def test_weapon() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.GI) as client:
        gi_new = await client.fetch_new()
        for weapon_id in gi_new.weapon_ids:
            await client.fetch_weapon_detail(weapon_id)


@pytest.mark.asyncio
async def test_artifact_sets() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.GI) as client:
        await client.fetch_artifact_sets()


@pytest.mark.asyncio
async def test_artifact_set() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.GI) as client:
        gi_new = await client.fetch_new()
        for artifact_set_id in gi_new.artifact_set_ids:
            await client.fetch_artifact_set_detail(artifact_set_id)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ..constants import ZZZ_LANG_MAP
from ..enums import Game, Language
from ..models import zzz
from .base import BaseClient

if TYPE_CHECKING:
    from collections.abc import Sequence

    from aiohttp import ClientSession

__all__ = ("ZZZClient",)


class ZZZClient(BaseClient):
    """Client to interact with the Hakushin Zenless Zone Zero API."""

    def __init__(
        self,
        lang: Language = Language.EN,
        *,
        cache_path: str = "./.cache/hakushin/aiohttp-cache.db",
        cache_ttl: int = 3600,
        headers: dict[str, Any] | None = None,
        debug: bool = False,
        session: ClientSession | None = None,
    ) -> None:
        super().__init__(
            Game.ZZZ,
            lang,
            cache_path=cache_path,
            cache_ttl=cache_ttl,
            headers=headers,
            debug=debug,
            session=session,
        )

    async def fetch_new(self, *, use_cache: bool = True) -> zzz.New:
        """Fetch the ID of beta items in Zenless Zone Zero.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A model representing the new items.
        """
        data = await self._request("new", use_cache, static=True)
        return zzz.New(**data)

    async def fetch_characters(self, *, use_cache: bool = True) -> list[zzz.Character]:
        """Fetch all Zenless Zone Zero characters.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of character objects.
        """
        data = await self._request("character", use_cache, in_data=True)

        characters = [
            zzz.Character(id=int(char_id), **char)
            for char_id, char in data.items()
            if char_id not in {"2011", "2021"}  # Exclude MCs
        ]
        for char in characters:
            char.name = char.names[ZZZ_LANG_MAP[self.lang]]

        return characters

    async def fetch_character_detail(
        self, character_id: int, *, use_cache: bool = True
    ) -> zzz.CharacterDetail:
        """Fetch the details of a Zenless Zone Zero character.

        Args:
            character_id: The character ID.
            use_cache: Whether to use the response cache.

        Returns:
            The character details object.
        """
        endpoint = f"character/{character_id}"
        data = await self._request(endpoint, use_cache)
        return zzz.CharacterDetail(**data)

    async def fetch_weapons(self, *, use_cache: bool = True) -> list[zzz.Weapon]:
        """Fetch all Zenless Zone Zero weapons (w-engines).

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of weapon objects.
        """
        data = await self._request("weapon", use_cache, in_data=True)

        weapons = [zzz.Weapon(id=int(weapon_id), **weapon) for weapon_id, weapon in data.items()]
        for weapon in weapons:
            weapon.name = weapon.names[ZZZ_LANG_MAP[self.lang]]

        return weapons

    async def fetch_weapon_detail(
        self, weapon_id: int, *, use_cache: bool = True
    ) -> zzz.WeaponDetail:
        """Fetch the details of a Zenless Zone Zero weapon.

        Args:
            weapon_id: The weapon ID.
            use_cache: Whether to use the response cache.

        Returns:
            The weapon details object.
        """
        endpoint = f"weapon/{weapon_id}"
        data = await self._request(endpoint, use_cache)
        return zzz.WeaponDetail(**data)

    async def fetch_bangboos(self, *, use_cache: bool = True) -> list[zzz.Bangboo]:
        """Fetch all Zenless Zone Zero bangboos.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of bangboo objects.
        """
        data = await self._request("bangboo", use_cache, in_data=True)
        bangboos = [
            zzz.Bangboo(id=int(bangboo_id), **bangboo) for bangboo_id, bangboo in data.items()
        ]
        for bangboo in bangboos:
            bangboo.name = bangboo.names[ZZZ_LANG_MAP[self.lang]]
        return bangboos

    async def fetch_bangboo_detail(
        self, bangboo_id: int, *, use_cache: bool = True
    ) -> zzz.BangbooDetail:
        """Fetch the details of a Zenless Zone Zero bangboo.

        Args:
            bangboo_id: The bangboo ID.
            use_cache: Whether to use the response cache.

        Returns:
            The bangboo details object.
        """
        endpoint = f"bangboo/{bangboo_id}"
        data = await self._request(endpoint, use_cache)
        return zzz.BangbooDetail(**data)

    async def fetch_drive_discs(self, *, use_cache: bool = True) -> list[zzz.DriveDisc]:
        """Fetch all Zenless Zone Zero drive discs.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of drive disc objects.
        """
        data = await self._request("equipment", use_cache, in_data=True)
        drive_discs = [
            zzz.DriveDisc(id=int(drive_disc_id), **drive_disc)
            for drive_disc_id, drive_disc in data.items()
        ]
        for drive_disc in drive_discs:
            if self.lang is Language.EN:
                info = drive_disc.en_info
            elif self.lang is Language.KO:
                info = drive_disc.ko_info
            elif self.lang is Language.ZH:
                info = drive_disc.chs_info
            else:
                info = drive_disc.ja_info

            if info is None:
                continue

            drive_disc.name = info.name
            drive_disc.two_piece_effect = info.two_piece_effect
            drive_disc.four_piece_effect = info.four_piece_effect

        return drive_discs

    async def fetch_drive_disc_detail(
        self, drive_disc_id: int, *, use_cache: bool = True
    ) -> zzz.DriveDiscDetail:
        """Fetch the details of a Zenless Zone Zero drive disc.

        Args:
            drive_disc_id: The drive disc ID.
            use_cache: Whether to use the response cache.

        Returns:
            The drive disc details object.
        """
        endpoint = f"equipment/{drive_disc_id}"
        data = await self._request(endpoint, use_cache)
        return zzz.DriveDiscDetail(**data)

    async def fetch_items(self, *, use_cache: bool = True) -> Sequence[zzz.Item]:
        """Fetch all Zenless Zone Zero items.

        Args:
            use_cache: Whether to use the response cache.

        Returns:
            A list of item objects.
        """
        data = await self._request("item", use_cache)
        items = [zzz.Item(id=int(item_id), **item) for item_id, item in data.items()]
        return items



from __future__ import annotations

from typing import Any, Literal

from pydantic import Field, field_validator, model_validator

from ...constants import GI_CHARA_RARITY_MAP
from ...enums import GIElement
from ..base import APIModel

__all__ = (
    "Character",
    "CharacterConstellation",
    "CharacterDetail",
    "CharacterInfo",
    "CharacterPassive",
    "CharacterSkill",
    "CharacterStatsModifier",
    "FightPropGrowthCurve",
    "Namecard",
    "SkillUpgradeInfo",
    "UpgradeMaterial",
    "UpgradeMaterialInfo",
    "UpgradeMaterialInfos",
)


class Namecard(APIModel):
    """Character's namecard."""

    id: int = Field(alias="Id")
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    icon: str = Field(alias="Icon")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"


class CharacterInfo(APIModel):
    """Character's information."""

    namecard: Namecard | None = Field(None, alias="Namecard")

    @field_validator("namecard", mode="before")
    def _handle_empty_namecard(cls, value: dict[str, Any] | None) -> dict[str, Any] | None:
        return value or None


class SkillUpgradeInfo(APIModel):
    """Character's skill upgrade information."""

    level: int = Field(alias="Level")
    icon: str = Field(alias="Icon")
    attributes: list[str] = Field(alias="Desc")
    parameters: list[float] = Field(alias="Param")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"

    @field_validator("attributes", mode="before")
    def _remove_empty_attributes(cls, value: list[str]) -> list[str]:
        return [attr for attr in value if attr]


class CharacterSkill(APIModel):
    """Character's skill."""

    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    upgrade_info: dict[str, SkillUpgradeInfo] = Field(alias="Promote")


class CharacterPassive(APIModel):
    """Character's passive talent."""

    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    unlock: int = Field(alias="Unlock")
    parameters: list[float] = Field(alias="ParamList")
    icon: str = Field(alias="Icon")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"


class CharacterConstellation(APIModel):
    """Character's constellation."""

    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    parameters: list[float] = Field(alias="ParamList")
    icon: str = Field(alias="Icon")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"


class UpgradeMaterial(APIModel):
    """Character's upgrade material."""

    name: str = Field(alias="Name")
    id: int = Field(alias="Id")
    count: int = Field(alias="Count")
    rarity: Literal[1, 2, 3, 4, 5] = Field(alias="Rank")

    @property
    def icon(self) -> str:
        """Material's icon URL."""
        return f"https://api.hakush.in/gi/UI/UI_ItemIcon_{self.id}.webp"


class UpgradeMaterialInfo(APIModel):
    """Character's upgrade material information."""

    materials: list[UpgradeMaterial] = Field(alias="Mats")
    mora_cost: int = Field(alias="Cost")


class UpgradeMaterialInfos(APIModel):
    """Character's upgrade material information."""

    ascensions: list[UpgradeMaterialInfo] = Field(alias="Ascensions")
    talents: list[list[UpgradeMaterialInfo]] = Field(alias="Talents")


class FightPropGrowthCurve(APIModel):
    """Character's stat growth curve data."""

    stat_type: str = Field(alias="type")
    growth_type: str = Field(alias="growCurve")


class CharacterStatsModifier(APIModel):
    """Character's stat modifiers."""

    hp: dict[str, float] = Field(alias="HP")
    atk: dict[str, float] = Field(alias="ATK")
    def_: dict[str, float] = Field(alias="DEF")
    ascension: list[dict[str, float]] = Field(alias="Ascension")
    prop_growth_curves: list[FightPropGrowthCurve] = Field(alias="PropGrowCurves")


class CharacterDetail(APIModel):
    """Genshin Impact character detail."""

    # Info
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    info: CharacterInfo = Field(alias="CharaInfo")
    rarity: Literal[4, 5] = Field(alias="Rarity")
    icon: str = Field(alias="Icon")

    # Combat
    skills: list[CharacterSkill] = Field(alias="Skills")
    passives: list[CharacterPassive] = Field(alias="Passives")
    constellations: list[CharacterConstellation] = Field(alias="Constellations")

    # Props
    stamina_recovery: float = Field(alias="StaminaRecovery")
    base_hp: float = Field(alias="BaseHP")
    base_atk: float = Field(alias="BaseATK")
    base_def: float = Field(alias="BaseDEF")
    crit_rate: float = Field(alias="CritRate")
    crit_dmg: float = Field(alias="CritDMG")

    stats_modifier: CharacterStatsModifier = Field(alias="StatsModifier")
    upgrade_materials: UpgradeMaterialInfos = Field(alias="Materials")

    @property
    def gacha_art(self) -> str:
        """Character's gacha art URL."""
        return self.icon.replace("AvatarIcon", "Gacha_AvatarImg")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"

    @field_validator("rarity", mode="before")
    def _convert_rarity(cls, value: str) -> Literal[4, 5]:
        return GI_CHARA_RARITY_MAP[value]


class Character(APIModel):
    """Genshin Impact character."""

    id: str  # This field is not present in the API response.
    icon: str
    rarity: Literal[4, 5] = Field(alias="rank")
    description: str = Field(alias="desc")
    element: GIElement | None = None
    names: dict[Literal["EN", "CHS", "KR", "JP"], str]
    name: str = Field(None)  # This value of this field is assigned in post processing.

    @field_validator("icon", mode="before")
    def _convert_icon(cls, value: str) -> str:
        return f"https://api.hakush.in/gi/UI/{value}.webp"

    @field_validator("rarity", mode="before")
    def _convert_rarity(cls, value: str) -> Literal[4, 5]:
        return GI_CHARA_RARITY_MAP[value]

    @field_validator("element", mode="before")
    def _convert_element(cls, value: str) -> GIElement | None:
        return GIElement(value) if value else None

    @model_validator(mode="before")
    def _transform_names(cls, values: dict[str, Any]) -> dict[str, Any]:
        # This is probably the most questionable API design decision I've ever seen.
        values["names"] = {
            "EN": values.pop("EN"),
            "CHS": values.pop("CHS"),
            "KR": values.pop("KR"),
            "JP": values.pop("JP"),
        }
        return values



from __future__ import annotations

import pytest

import hakushin


@pytest.mark.asyncio
async def test_new() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
        await client.fetch_new()


@pytest.mark.asyncio
async def test_characters() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
        await client.fetch_characters()


@pytest.mark.asyncio
async def test_character_detail() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
        characters = await client.fetch_characters()
        for ch in characters:
            await client.fetch_character_detail(ch.id)


@pytest.mark.asyncio
async def test_weapons() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
        await client.fetch_weapons()


@pytest.mark.asyncio
async def test_weapon_detail() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
        weapons = await client.fetch_weapons()
        for weapon in weapons:
            await client.fetch_weapon_detail(weapon.id)


@pytest.mark.asyncio
async def test_bangbooss() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
        await client.fetch_bangboos()


@pytest.mark.asyncio
async def test_bangboo_detail() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
        bangboos = await client.fetch_bangboos()
        for bangboo in bangboos:
            await client.fetch_bangboo_detail(bangboo.id)


@pytest.mark.asyncio
async def test_drive_discs() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
        await client.fetch_drive_discs()


@pytest.mark.asyncio
async def test_drive_disc_detail() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
        drive_discs = await client.fetch_drive_discs()
        for drive_disc in drive_discs:
            await client.fetch_drive_disc_detail(drive_disc.id)


@pytest.mark.asyncio
async def test_items() -> None:
    async with hakushin.HakushinAPI(hakushin.Game.ZZZ) as client:
        await client.fetch_items()



from __future__ import annotations

from typing import Any, Literal

from pydantic import Field, field_validator, model_validator

from ...constants import ZZZ_SA_RARITY_CONVERTER
from ...utils import cleanup_text
from ..base import APIModel
from .common import ZZZExtraProp, ZZZMaterial

__all__ = ("Bangboo", "BangbooAscension", "BangbooDetail", "BangbooSkill")


class Bangboo(APIModel):
    """ZZZ bangboo model."""

    id: int
    icon: str
    rarity: Literal["S", "A"] | None = Field(alias="rank")
    code_name: str = Field(alias="codename")
    description: str = Field(alias="desc")
    name: str = Field(None)  # This field doesn't exist in the API response
    names: dict[Literal["EN", "JA", "CHS", "KO"], str]

    @field_validator("icon")
    @classmethod
    def __convert_icon(cls, value: str) -> str:
        value = value.split("/")[-1].split(".")[0]
        return f"https://api.hakush.in/zzz/UI/{value}.webp"

    @field_validator("rarity", mode="before")
    @classmethod
    def __convert_rarity(cls, value: int | None) -> Literal["S", "A"] | None:
        return ZZZ_SA_RARITY_CONVERTER[value] if value is not None else None

    @model_validator(mode="before")
    @classmethod
    def __pop_names(cls, values: dict[str, Any]) -> dict[str, Any]:
        values["names"] = {
            "EN": values.pop("EN"),
            "KO": values.pop("KO"),
            "CHS": values.pop("CHS"),
            "JA": values.pop("JA"),
        }
        return values


class BangbooAscension(APIModel):
    """ZZZ bangboo ascension model."""

    max_hp: int = Field(alias="HpMax")
    attack: int = Field(alias="Attack")
    defense: int = Field(alias="Defence")
    max_level: int = Field(alias="LevelMax")
    min_level: int = Field(alias="LevelMin")
    materials: list[ZZZMaterial] = Field(alias="Materials")
    extra_props: list[ZZZExtraProp] = Field(alias="Extra")

    @field_validator("extra_props", mode="before")
    @classmethod
    def __convert_extra_props(cls, value: dict[str, dict[str, Any]]) -> list[ZZZExtraProp]:
        return [ZZZExtraProp(**prop) for prop in value.values()]

    @field_validator("materials", mode="before")
    @classmethod
    def __convert_materials(cls, value: dict[str, int]) -> list[ZZZMaterial]:
        return [ZZZMaterial(id=int(id_), amount=amount) for id_, amount in value.items()]


class BangbooSkill(APIModel):
    """ZZZ bangboo skill model."""

    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    properties: list[str] = Field(alias="Property")
    parameter: str = Field(alias="Param")

    @field_validator("description")
    @classmethod
    def __cleanup_text(cls, value: str) -> str:
        return cleanup_text(value)


class BangbooDetail(APIModel):
    """ZZZ bangboo detail model."""

    id: int = Field(alias="Id")
    code_name: str = Field(alias="CodeName")
    name: str = Field(alias="Name")
    description: str = Field(alias="Desc")
    rarity: Literal["S", "A"] = Field(alias="Rarity")
    icon: str = Field(alias="Icon")
    stats: dict[str, float] = Field(alias="Stats")
    ascensions: dict[str, BangbooAscension] = Field(alias="Level")
    """Dictionary of ascension objects, key starts from 1."""
    skills: dict[Literal["A", "B", "C"], dict[str, BangbooSkill]] = Field(alias="Skill")

    @field_validator("skills", mode="before")
    @classmethod
    def __unnest_level(
        cls, value: dict[Literal["A", "B", "C"], dict[Literal["Level"], dict[str, Any]]]
    ) -> dict[Literal["A", "B", "C"], dict[str, BangbooSkill]]:
        return {key: value[key]["Level"] for key in value}

    @field_validator("icon")
    @classmethod
    def __convert_icon(cls, value: str) -> str:
        value = value.split("/")[-1].split(".")[0]
        return f"https://api.hakush.in/zzz/UI/{value}.webp"

    @field_validator("rarity", mode="before")
    @classmethod
    def __convert_rarity(cls, value: int) -> Literal["S", "A"]:
        return ZZZ_SA_RARITY_CONVERTER[value]



from __future__ import annotations

from typing import Final, Literal

from .enums import Game, HSRPath, Language

__all__ = (
    "GI_CHARA_RARITY_MAP",
    "GI_LANG_MAP",
    "HSR_API_LANG_MAP",
    "HSR_CHARA_RARITY_MAP",
    "HSR_LIGHT_CONE_RARITY_MAP",
    "PERCENTAGE_FIGHT_PROPS",
)

GI_CHARA_RARITY_MAP: Final[dict[str, Literal[4, 5]]] = {
    "QUALITY_PURPLE": 4,
    "QUALITY_ORANGE": 5,
    "QUALITY_ORANGE_SP": 5,
}

HSR_CHARA_RARITY_MAP: Final[dict[str, Literal[4, 5]]] = {
    "CombatPowerAvatarRarityType4": 4,
    "CombatPowerAvatarRarityType5": 5,
}

HSR_LIGHT_CONE_RARITY_MAP: Final[dict[str, Literal[3, 4, 5]]] = {
    "CombatPowerLightconeRarity3": 3,
    "CombatPowerLightconeRarity4": 4,
    "CombatPowerLightconeRarity5": 5,
}

HSR_API_LANG_MAP: Final[dict[Language, Literal["en", "jp", "kr", "cn"]]] = {
    Language.EN: "en",
    Language.JA: "jp",
    Language.KO: "kr",
    Language.ZH: "cn",
}
"""Map to convert API language enum to HSR API language."""

GI_LANG_MAP: Final[dict[Language, Literal["EN", "JP", "KR", "CHS"]]] = {
    Language.EN: "EN",
    Language.JA: "JP",
    Language.KO: "KR",
    Language.ZH: "CHS",
}
"""Map to convert API language enum to GI data language."""

ZZZ_LANG_MAP: Final[dict[Language, Literal["EN", "KO", "CHS", "JA"]]] = {
    Language.EN: "EN",
    Language.JA: "JA",
    Language.KO: "KO",
    Language.ZH: "CHS",
}

PERCENTAGE_FIGHT_PROPS: Final[set[str]] = {
    "FIGHT_PROP_HP_PERCENT",
    "FIGHT_PROP_ATTACK_PERCENT",
    "FIGHT_PROP_DEFENSE_PERCENT",
    "FIGHT_PROP_SPEED_PERCENT",
    "FIGHT_PROP_CRITICAL",
    "FIGHT_PROP_CRITICAL_HURT",
    "FIGHT_PROP_CHARGE_EFFICIENCY",
    "FIGHT_PROP_ADD_HURT",
    "FIGHT_PROP_HEAL_ADD",
    "FIGHT_PROP_HEALED_ADD",
    "FIGHT_PROP_FIRE_ADD_HURT",
    "FIGHT_PROP_WATER_ADD_HURT",
    "FIGHT_PROP_GRASS_ADD_HURT",
    "FIGHT_PROP_ELEC_ADD_HURT",
    "FIGHT_PROP_ICE_ADD_HURT",
    "FIGHT_PROP_WIND_ADD_HURT",
    "FIGHT_PROP_PHYSICAL_ADD_HURT",
    "FIGHT_PROP_ROCK_ADD_HURT",
    "FIGHT_PROP_SKILL_CD_MINUS_RATIO",
    "FIGHT_PROP_ATTACK_PERCENT_A",
    "FIGHT_PROP_DEFENSE_PERCENT_A",
    "FIGHT_PROP_HP_PERCENT_A",
    "criticalChance",
    "criticalDamage",
    "breakDamageAddedRatio",
    "breakDamageAddedRatioBase",
    "healRatio",
    "sPRatio",
    "statusProbability",
    "statusResistance",
    "criticalChanceBase",
    "criticalDamageBase",
    "healRatioBase",
    "sPRatioBase",
    "statusProbabilityBase",
    "statusResistanceBase",
    "physicalAddedRatio",
    "physicalResistance",
    "fireAddedRatio",
    "fireResistance",
    "iceAddedRatio",
    "iceResistance",
    "thunderAddedRatio",
    "thunderResistance",
    "windAddedRatio",
    "windResistance",
    "quantumAddedRatio",
    "quantumResistance",
    "imaginaryAddedRatio",
    "imaginaryResistance",
    "hPAddedRatio",
    "attackAddedRatio",
    "defenceAddedRatio",
    "healTakenRatio",
    "physicalResistanceDelta",
    "fireResistanceDelta",
    "iceResistanceDelta",
    "thunderResistanceDelta",
    "windResistanceDelta",
    "quantumResistanceDelta",
    "imaginaryResistanceDelta",
}
"""Set of fight props that should be displayed as percentage value."""

HSR_PATH_NAMES: Final[dict[Language, dict[HSRPath, str]]] = {
    Language.EN: {
        HSRPath.ABUNDANCE: "Abundance",
        HSRPath.DESTRUCTION: "Destruction",
        HSRPath.ERUDITION: "Erudition",
        HSRPath.HARMONY: "Harmony",
        HSRPath.NIHILITY: "Nihility",
        HSRPath.PRESERVATION: "Preservation",
        HSRPath.THE_HUNT: "The Hunt",
    },
    Language.JA: {
        HSRPath.ABUNDANCE: "豊穣",
        HSRPath.DESTRUCTION: "壊滅",
        HSRPath.ERUDITION: "知恵",
        HSRPath.HARMONY: "調和",
        HSRPath.NIHILITY: "虚無",
        HSRPath.PRESERVATION: "存護",
        HSRPath.THE_HUNT: "巡狩",
    },
    Language.ZH: {
        HSRPath.ABUNDANCE: "豐饒",
        HSRPath.DESTRUCTION: "毀滅",
        HSRPath.ERUDITION: "智識",
        HSRPath.HARMONY: "同諧",
        HSRPath.NIHILITY: "虚無",
        HSRPath.PRESERVATION: "存護",
        HSRPath.THE_HUNT: "巡獵",
    },
    Language.KO: {
        HSRPath.ABUNDANCE: "풍요",
        HSRPath.DESTRUCTION: "파멸",
        HSRPath.ERUDITION: "지식",
        HSRPath.HARMONY: "화합",
        HSRPath.NIHILITY: "공허",
        HSRPath.PRESERVATION: "보존",
        HSRPath.THE_HUNT: "수렵",
    },
}

TRAILBLAZER_NAMES: Final[dict[Language, str]] = {
    Language.EN: "Trailblazer",
    Language.JA: "開拓者",
    Language.ZH: "开拓者",
    Language.KO: "개척자",
}

STAT_TO_FIGHT_PROP: Final[dict[str, str]] = {
    "BaseHP": "FIGHT_PROP_BASE_HP",
    "BaseDEF": "FIGHT_PROP_BASE_DEFENSE",
    "BaseATK": "FIGHT_PROP_BASE_ATTACK",
}

NOT_ASCENDED_LEVEL_TO_ASCENSION: Final[dict[Game, dict[int, int]]] = {
    Game.GI: {80: 5, 70: 4, 60: 3, 50: 2, 40: 1, 20: 0},
    Game.HSR: {70: 5, 60: 4, 50: 3, 40: 2, 30: 1, 20: 0},
}

ASCENDED_LEVEL_TO_ASCENSION: Final[dict[Game, dict[tuple[int, int], int]]] = {
    Game.GI: {(80, 90): 6, (70, 80): 5, (60, 70): 4, (50, 60): 3, (40, 50): 2, (20, 40): 1},
    Game.HSR: {(70, 80): 6, (60, 70): 5, (50, 60): 4, (40, 50): 3, (30, 40): 2, (20, 30): 1},
}

ASCENSION_TO_MAX_LEVEL: Final[dict[Game, dict[int, int]]] = {
    Game.GI: {0: 20, 1: 40, 2: 50, 3: 60, 4: 70, 5: 80, 6: 90},
    Game.HSR: {0: 20, 1: 30, 2: 40, 3: 50, 4: 60, 5: 70, 6: 80},
}

ZZZ_SAB_RARITY_CONVERTER: Final[dict[int, Literal["B", "A", "S"]]] = {2: "B", 3: "A", 4: "S"}
ZZZ_SA_RARITY_CONVERTER: Final[dict[int, Literal["A", "S"]]] = {3: "A", 4: "S"}
